import json


class DataStore:
    def __init__(self):
        self._value: list[dict[str, str]] = []

    @staticmethod
    def data_unit(title: str, id: str) -> dict[str, str]:
        return {'title': title, 'id': id}

    def update(self, title: str, id: str):
        value = self.data_unit(title, id)
        if value not in self._value:
            self._value.append(value)

    def list(self) -> list[dict[str, str]]:
        return self._value

    def purge(self) -> None:
        self._value = []

    def print(self) -> None:
        if len(self) > 0:
            for i, item in enumerate(self.list(), start=1):
                print(i, '-', item['title'])
        else:
            raise Exception('Datastore is empty')

    @classmethod
    def retrieve_from_file(cls, file_fullpath):
        new_data_units = []

        def retrieve(data_units) -> None:
            if type(data_units) == dict:
                try:
                    title, id = data_units.keys()
                    if title == 'title' and id == 'id':
                        new_data_units.append(data_units)
                except Exception:
                    pass
            elif type(data_units) == list:
                for data_unit in data_units:
                    retrieve(data_unit)

        try:
            with open(file_fullpath) as f:
                try:
                    file = json.load(f)
                    retrieve(file)
                except Exception:
                    pass
        except Exception:
            pass
        data_store = cls()
        data_store._value = new_data_units
        return data_store

    def append(self, data_store) -> None:
        """
        Append another DataStore into this DataStore.
        Use with caution, value added by load() can cause duplication.
        """
        if type(data_store) != DataStore:
            raise TypeError(
                f"Only {self.__class__.__name__} object can be loaded")
        self._value += data_store.list()

    def __getitem__(self, index: int) -> dict[str, str]:
        return self._value[index]

    def __len__(self) -> int:
        return len(self._value)

    def __str__(self) -> str:
        return f'{self.__class__.__name__}{self.list()}'



import os
import json

from ytsort.data_store import DataStore


class CacheManager:
    """
    Provide functions to easily manage cache.
    Use json to dump data stored in DataStore object as list.
    """

    def __init__(self, config: dict) -> None:
        """
        Parameters
        ----------
        config: dict
            Configurations to change the behaviour of program

        Required keys in config:
        1. 'local_cache'
        2. 'shared_cache'
        """
        self.__local_playlist_cache_path = os.path.join(
            config['local_cache'], 'playlist.json')
        self.__local_channel_cache_path = os.path.join(
            config['local_cache'], 'channel.json')
        self.__shared_channel_cache_path = os.path.join(
            config['shared_cache'], 'channel.json')

        self.__indent = config['cache_indent']

        self._local_playlist_cache = DataStore()
        self._local_channel_cache = DataStore()
        self._shared_channel_cache = DataStore()

        self.create_folder(config['local_cache'])
        self.create_folder(config['shared_cache'])

        self.load()

    @property
    def local_playlist_cache(self):
        """Playlist cache stored in renaming folder"""
        return self._local_playlist_cache

    @property
    def local_channel_cache(self):
        """Channel cache stored in renaming folder"""
        return self._local_channel_cache

    @property
    def shared_channel_cache(self):
        """Channel cache stored in shared cache folder"""
        return self._shared_channel_cache

    @staticmethod
    def _create_cache(cache_unit: dict):
        local_cache = DataStore()
        local_cache.update(
            cache_unit['title'], cache_unit['id'])
        return local_cache

    def update_playlist_cache(self, playlist_cache_unit: dict):
        self.update_local_playlist_cache(playlist_cache_unit)

    def update_channel_cache(self, channel_cache_unit: dict):
        self.update_local_channel_cache(channel_cache_unit)
        self.update_shared_channel_cache(channel_cache_unit)

    def update_local_playlist_cache(self, playlist_cache_unit: dict):
        self._local_playlist_cache = self._create_cache(playlist_cache_unit)

    def update_local_channel_cache(self, channel_cache_unit: dict):
        self._local_channel_cache = self._create_cache(channel_cache_unit)

    def update_shared_channel_cache(self, channel_cache_unit: dict):
        self._shared_channel_cache.update(
            channel_cache_unit['title'], channel_cache_unit['id'])

    def load(self):
        self.load_local_playlist_cache()
        self.load_local_channel_cache()
        self.load_shared_channel_cache()

    def load_local_playlist_cache(self):
        r = DataStore.retrieve_from_file(self.__local_playlist_cache_path)
        if r:
            self._local_playlist_cache.append(r)

    def load_local_channel_cache(self):
        r = DataStore.retrieve_from_file(self.__local_channel_cache_path)
        if r:
            self._local_channel_cache.append(r)

    def load_shared_channel_cache(self):
        r = DataStore.retrieve_from_file(self.__shared_channel_cache_path)
        if r:
            self._shared_channel_cache.append(r)

    @staticmethod
    def create_folder(folder):
        if not os.path.exists(folder):
            os.mkdir(folder)

    def is_local_playlist_cache_available(self):
        return True if len(self._local_playlist_cache) > 0 else False

    def is_local_channel_cache_available(self):
        return True if len(self._local_channel_cache) > 0 else False

    def is_shared_channel_cache_available(self):
        return True if len(self._shared_channel_cache) > 0 else False

    def _dump(self, path, cache_list: list):
        with open(path, 'w') as f:
            json.dump(cache_list, f, indent=self.__indent)

    def save(self):
        self.save_local_playlist_cache()
        self.save_local_channel_cache()
        self.save_shared_channel_cache()

    def save_local_playlist_cache(self):
        self._dump(self.__local_playlist_cache_path,
                   self.local_playlist_cache.list())

    def save_local_channel_cache(self):
        self._dump(self.__local_channel_cache_path,
                   self.local_channel_cache.list())

    def save_shared_channel_cache(self):
        self._dump(self.__shared_channel_cache_path,
                   self.shared_channel_cache.list())

    def delete_local_playlist_cache(self):
        self._local_playlist_cache.purge()
        try:
            os.remove(self.__local_playlist_cache_path)
        except Exception:
            pass

    def delete_local_channel_cache(self):
        self._local_channel_cache.purge()
        try:
            os.remove(self.__local_channel_cache_path)
        except Exception:
            pass



# These are defaults. 'character_after_serial' and 'padded_zero'
# can be changed from command line options. 'api_key' can be provided while running.

import os

config = {
    'api_key': os.environ.get('YOUTUBE_DATA_API_KEY'),
    'exceptions_file': '.ytsignore',
    'local_cache': os.path.join(os.getcwd(), '.ytsort_cache/'),
    'shared_cache': os.path.join(os.path.dirname(__file__), '.ytsort_cache/'),
    'cache_indent': 4,
    'character_after_serial': ')',
    'padded_zero': False,
}



from googleapiclient.discovery import Resource
from threading import Thread
from copy import deepcopy
from alive_progress import alive_bar
from math import ceil

from ytsort.my_io import capture_stdout, print_heading
from ytsort.data_store import DataStore


class Youtube:
    """
    Access YouTube level data, i.e. search channels and subscribers.
    """

    def __init__(self, yt_resource: Resource) -> None:
        """
        Parameters
        ----------
        yt_resource:
            googleapiclient.discovery.build object
        """
        self._yt_resource = yt_resource
        self._channel_store = DataStore()
        self._subscribers = []

    def channels(self) -> DataStore:
        return self._channel_store

    def search_channel(self, query: str) -> None:
        yt_request = self._yt_resource.search().list(
            part='snippet',
            type='channel',
            maxResults=9,
            q=query
        )
        yt_response = yt_request.execute()

        for item in yt_response['items']:
            self._channel_store.update(
                title=item['snippet']['title'], id=item['snippet']['channelId']
            )

    def print_channels(self):
        channel_output = capture_stdout(
            lambda: self._channel_store.print())
        self._fetch_subscribers_of_all()

        print_heading('#   Channel (Subscribers)')
        for index, out in enumerate(channel_output):
            print(f'{out} ({self._subscribers[index]})')

    def subscriber_count(self, channel_id):
        def shorten_M_K(subs: str):
            length = len(str(subs))
            subs = int(subs)
            if length > 6:
                subs /= 10**6
                subs = str(subs)
                if length >= 9:
                    subs = subs[:length-6]+'M'
                else:
                    subs = subs[:4]+'M'
            elif length > 3:
                subs /= 10**3
                subs = str(subs)
                if length == 6:
                    subs = subs[:3]+'K'
                else:
                    subs = subs[:4]+'K'
            return subs

        # self._yt_resource is not working if used directly in multi-threading. Problem not known.
        # Error: malloc(): unsorted double linked list corrupted
        # Using deepcopy it works fine
        yt_response = deepcopy(self._yt_resource).channels().list(
            part='statistics',
            id=channel_id,
        ).execute()

        if yt_response['items'][0]['statistics']['hiddenSubscriberCount']:
            return
        return shorten_M_K(yt_response['items'][0]['statistics']['subscriberCount'])

    def _fetch_subscribers_of_all(self):
        def get_subs(channel_id, index, progress_bar):
            subs = self.subscriber_count(channel_id)
            if not subs:
                subs = 'Hidden'
            self._subscribers[index] = subs
            progress_bar()

        self._subscribers = [None]*len(self._channel_store)

        with alive_bar(total=len(self._subscribers), bar='smooth', spinner='dots_reverse') as bar:
            threads = []
            for index, channel_cache_unit in enumerate(self._channel_store.list()):
                thread = Thread(target=get_subs, args=(
                    channel_cache_unit['id'], index, bar))
                threads.append(thread)
                thread.start()

            for thread in threads:
                thread.join()


class YTChannel:
    """
    Single YouTube channel.
    Contains channel's ID and all the playlists available in the channel.
    """

    def __init__(self, yt_resource: Resource, channel_id) -> None:
        """
        Parameters
        ----------
        yt_resource:
            googleapiclient.discovery.Resource object
        channel_id:
            youtube channel ID
        """
        self._yt_resource = yt_resource
        self._channel_id = channel_id
        self._playlist_store = DataStore()

    def fetch_playlists(self):
        with alive_bar(manual=True, bar='smooth', spinner='dots_reverse') as bar:

            nextPageToken = None
            page = 1
            while True:
                pl_request = self._yt_resource.playlists().list(
                    part='snippet',
                    channelId=self._channel_id,
                    pageToken=nextPageToken
                )
                pl_response = pl_request.execute()

                for item in pl_response['items']:
                    self._playlist_store.update(
                        title=item['snippet']['title'], id=item['id']
                    )

                bar(page / ceil(pl_response['pageInfo']['totalResults'] /
                                pl_response['pageInfo']['resultsPerPage']))

                page += 1

                nextPageToken = pl_response.get('nextPageToken')
                if not nextPageToken:
                    break
        print()

    def playlists(self) -> DataStore:
        return self._playlist_store


class YTPlaylist:
    """
    YouTube channel's single playlist.
    Contains playlist's ID and all the videos available in the playlist.
    """

    def __init__(self, yt_resource: Resource, playlist_id) -> None:
        """
        Parameters
        ----------
        yt_resource:
            googleapiclient.discovery.Resource object
        playlist_id:
            youtube playlist ID
        """
        self._yt_resource = yt_resource
        self._playlist_id = playlist_id
        self._videos_store = DataStore()

    def fetch_videos(self):
        with alive_bar(manual=True, bar='smooth', spinner='dots_reverse') as bar:

            nextPageToken = None
            page = 1
            while True:
                pl_item_request = self._yt_resource.playlistItems().list(
                    part='snippet',
                    playlistId=self._playlist_id,
                    pageToken=nextPageToken
                )
                pl_item_response = pl_item_request.execute()

                for item in pl_item_response['items']:
                    self._videos_store.update(
                        title=item['snippet']['title'], id='#'
                    )

                bar(page / ceil(pl_item_response['pageInfo']['totalResults'] /
                                pl_item_response['pageInfo']['resultsPerPage']))
                page += 1

                nextPageToken = pl_item_response.get('nextPageToken')
                if not nextPageToken:
                    break
        print()

    def videos(self) -> DataStore:
        return self._videos_store

    def get_videos_serial(self) -> dict[str, int]:
        # {Title: serial_number}
        video_serial: dict[str, int] = {}
        for i, cache_unit in enumerate(self._videos_store.list(), start=1):
            video_serial.update({cache_unit['title']: i})
        return video_serial



import contextlib
import io
from getpass import getpass
from typing import Callable
from colorama import Fore as c, Style


def is_int(num: str) -> bool:
    """Check if given string is integer or not"""
    try:
        int(num)
    except ValueError:
        return False
    return True


def arg_parse(*args) -> str:
    """Take arguments and return after joining them by a space."""
    return ' '.join([str(x) for x in args])


def take_input(arg):
    arg = c.LIGHTGREEN_EX + arg + c.RESET
    return input(arg)


def non_empty_input(arg):
    """Takes input until the input is not empty."""
    arg = c.LIGHTGREEN_EX + arg + c.RESET
    while True:
        inp = input(arg)
        if inp != '':
            return inp


def non_empty_getpass(arg) -> str:
    """Takes hidden input until the input is not empty."""
    arg = c.LIGHTGREEN_EX + arg + c.RESET
    while True:
        inp = getpass(arg)
        if inp != '':
            return inp


def input_in_range(msg, a, b=None) -> int:
    """
    Takes input until the input is not in range. Both `a` and `b` inclusive.

    Parameters
    ----------
    a: int
        Upper bound if `b` is not supplied, else lower bound
    b: int
        Upper bound if supplied

    If only `a` is supplied, 0 is lower bound and `a` is upper bound.
    If both `a` and `b` are supplied, `a` is lower bound and `b` is upper bound.
    """
    if b is not None:
        if b <= a:
            raise ValueError('Upper bound should be greater than lower bound')
        lb = a
        ub = b
    else:
        if a <= 0:
            raise ValueError('Upper bound should be greater than 0')
        lb = 0
        ub = a

    while True:
        inp = non_empty_input(msg)
        try:
            inp = int(inp)
            if lb <= inp <= ub:
                return inp
            print_warning(f'Warning: Input range is [{lb},{ub}]')
        except ValueError:
            print_warning(f'Warning: Please input a number')


def print_info(*args):
    print(c.LIGHTYELLOW_EX + arg_parse(*args) + c.RESET)


def print_warning(*args):
    print(c.LIGHTRED_EX + arg_parse(*args) + c.RESET)


def print_heading(*args):
    print()
    print(Style.BRIGHT+c.LIGHTBLUE_EX, end='')
    st = arg_parse(*args)
    print(st)
    print('-'*len(st), end='')
    print(c.RESET + Style.NORMAL)


def capture_stdout(func: Callable) -> list:
    """
    Captures the standard output on running `func()` and returns it in a list with each line as its item.
    """
    f = io.StringIO()
    with contextlib.redirect_stdout(f):
        func()
    output = f.getvalue()
    output = output.split('\n')
    output.remove('')
    return output



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ytsort']

package_data = \
{'': ['*']}

install_requires = \
['alive-progress==1.6.2',
 'click>=8.0.3,<9.0.0',
 'colorama>=0.3.7,<0.4.0',
 'google-api-python-client>=2.8.0,<3.0.0']

entry_points = \
{'console_scripts': ['ytsort = ytsort.__main__:cli']}

setup_kwargs = {
    'name': 'ytsort',
    'version': '1.2.2',
    'description': 'Arrange downloaded youtube videos',
    'long_description': "# YTSort\n\nThis program sorts the already downloaded youtube videos present in a folder by renaming them and adding serial number before their names.\n\n# Requirements\n\nYoutube Data API v3 is required. Get it from [here](https://console.cloud.google.com/apis/library/youtube.googleapis.com?supportedpurview=project)\n\n# Install\n###### Recommended (To install pipx click [here](https://github.com/pypa/pipx#install-pipx))\n```\npipx install ytsort\n```\n\n###### or\n```\npip install ytsort\n```\n\n#### Or upgrade by:\n```\npipx upgrade ytsort\n```\n###### or\n```\npip install --upgrade ytsort\n```\n# Usage\n\nSet Youtube API Key to the environment variable 'YOUTUBE_DATA_API_KEY' (for ease of use, but not required).\n\n### Execute:\n```\nytsort\n```\n\n```\nUsage: ytsort [OPTIONS]\n\nOptions:\n\n  -c, --character TEXT    Character after serial.\n  -z, --zero              Add zero before serial numbers to make them all of\n                          equal length.\n  -x, --nozero            Don't add zero before serial numbers.\n  --help                  Show this message and exit.\n```\n\nIf -z & -x both are passed simultaneously then default from config.py will be used.\n\n\n# Install from source\nPoetry is required. For installation click [here](https://python-poetry.org/docs/#installation).\n\n1. Download the source and install the dependencies by running:\n  \n   ``` \n   git clone https://github.com/aqdasak/YTSort.git\n   cd YTSort\n   poetry install\n   ```\n\n2. Not required but for ease of use\n \n   a) Set Youtube API Key to the environment variable 'YOUTUBE_DATA_API_KEY'\n\n   or\n \n   b) edit the `config.py`:\n\n      `'api_key': os.environ.get('YOUTUBE_DATA_API_KEY'),` to `'api_key': <Your Youtube API key>,`\n\n### Run\nIn the source folder containing pyproject.toml\n```\npoetry shell\n```\n\nthen cd into the folder containing youtube videos and execute:\n```\nytsort\n```\n",
    'author': 'Aqdas Ahmad Khan',
    'author_email': 'aqdasak@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/aqdasak/YTSort/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import click
import os
from googleapiclient.discovery import build, Resource

from ytsort.config import config
from ytsort.data_store import DataStore
from ytsort.my_io import input_in_range, is_int, non_empty_getpass, print_heading, print_info, non_empty_input, print_warning, take_input
from ytsort.renamer import Renamer
from ytsort.cache_manager import CacheManager
from ytsort.youtube import YTPlaylist, Youtube, YTChannel


def get_exceptions() -> list[str]:
    """List of files not to be renamed"""

    exceptions = []
    try:
        with open(config['exceptions_file']) as f:
            exceptions.extend(f.read().split("\n"))
    except Exception:
        pass

    exceptions = [i for i in exceptions if i != '']
    if len(exceptions) > 0:
        print('\nFollowing files will be ignored:')
        for i, file in enumerate(exceptions, start=1):
            print(f"{i}. '{file}'")
    else:
        print('\nNo file ignored.')

    exceptions.extend([config['exceptions_file'], config['local_cache']])
    return exceptions


def print_and_select_playlist(playlists: DataStore) -> dict[str, str]:
    """
    Print the given playlists and take user input to select from the available playlists.
    If user inputs a number the playlist corresponding to that serial number will be selected.
    If user inputs a string, it will select only those playlists which contains the input as the substring of the playlist. The function will be run recursively with only those playlists.

    Parameters
    ----------
    playlists: DataStore
        Collection of playlists from which a playlist is selected by user

    Return
    ------
    : dict[str,str]
        Dictionary of type {'title':title,'id':id}
    """

    def p_a_s_p(original_playlists: DataStore, current_playlists: DataStore, warning: str = '') -> dict[str, str]:
        """
        Helper function

        Parameters
        ----------
        warning: str (Optional)
            Give warning if previous input was wrong. `warning` contains the warning message
        """

        if warning:
            print_warning(warning)
        else:
            current_playlists.print()

        if len(current_playlists) < len(original_playlists):
            print_info("Enter '/' to reset search")

        user_input = non_empty_input(
            'Select playlist by number or search name: ').lower()
        print()

        if is_int(user_input):
            if 1 <= int(user_input) <= len(current_playlists):
                # Select and return the playlist if user_input is in correct range
                return current_playlists[int(user_input)-1]

            # Display warning if user_input is out of range and pass the playlists unchanged
            return p_a_s_p(original_playlists, current_playlists, warning=f'Wrong input: Input range is [1,{len(current_playlists)}]')

        elif user_input == '/':
            # Reset the search
            return p_a_s_p(original_playlists, original_playlists)

        else:
            # Select playlists containing user_input as their title's substring
            temp_playlist = DataStore()
            for playlist in current_playlists:
                title = playlist['title']
                if user_input in title.lower():
                    temp_playlist.update(title, playlist['id'])

            # Rerun the function to take input on matched playlists
            if len(temp_playlist) > 0:
                return p_a_s_p(original_playlists, temp_playlist)

        # Display warning if no playlist is matched (and all other cases, if happen) and pass the playlists unchanged
        return p_a_s_p(original_playlists, current_playlists, warning="Didn't match, try again")

    return p_a_s_p(playlists, playlists)


def get_playlist_id_from_local_cache(cache: CacheManager) -> str | None:
    """
    Fetch playlist cache from local cache and ask user if the playlist is correct or not. If the playlist is correct return playlist's ID else delete local cache and return None.

    Parameters
    ----------
    cache: CacheManager
        CacheManager object

    Returns
    -------
        playlist_id: str
            ID of the YouTube playlist
        None
            If ID is not discarded by user 
    """

    playlist_cache_unit = cache.local_playlist_cache.list()[0]
    print_info(
        f"\nPlaylist: \"{playlist_cache_unit['title']}\" found in cache.")

    inp = take_input('Is the playlist correct?(n if no): ').lower()
    if inp in ('n', 'no'):
        cache.delete_local_playlist_cache()
        print()
        return None

    return playlist_cache_unit['id']


def get_playlist_id_from_youtube(cache: CacheManager, yt_resource: Resource) -> str:
    """
    Fetch playlists from YouTube and return playlist's ID

    Parameters
    ----------
    cache: CacheManager
        CacheManager object
    yt_resource: googleapiclient.discovery.Resource
        Required to access YouTube's Data API

    Returns
    -------
        playlist_id: str
            ID of the YouTube playlist
    """

    channel_id = get_channel_id(cache, yt_resource)
    channel = YTChannel(yt_resource, channel_id)
    channel.fetch_playlists()

    print_heading('<< PLAYLISTS >>')
    playlist_cache_unit = print_and_select_playlist(channel.playlists())

    cache.update_playlist_cache(playlist_cache_unit)
    return playlist_cache_unit['id']


def get_channel_id(cache: CacheManager, yt_resource: Resource) -> str:
    """
    Check Channel ID in local cache, if not found then in shared cache, and then search on YouTube and then returns the Channel ID.

    Parameters
    ----------
    cache: CacheManager
        CacheManager object to manage cache
    yt_resource: googleapiclient.discovery.Resource
        Required to access YouTube's Data API

    Returns
    -------
        channel_id: str
            ID of the YouTube channel
    """

    def get_channel_from_youtube(search_query: str, yt_resource: Resource) -> dict[str, str]:
        """
        Search channels on YouTube, print them, ask user to select among them and return the user selected channel

        Parameters
        ----------
        search_query: str
            Channel name to be searched on YouTube
        yt_resource: googleapiclient.discovery.Resource
            Required to access YouTube's Data API

        Returns
        -------
            channel: dict[str, str]
                YouTube channel dictionary {'title': <title>, 'id': <id>}
        """

        youtube = Youtube(yt_resource)
        while True:
            youtube.search_channel(search_query)
            if len(youtube.channels()) > 0:
                break
            search_query = non_empty_input('Not found search again: ')

        youtube.print_channels()
        select = int(input_in_range(
            'Select channel: ', 1, len(youtube.channels())))
        return youtube.channels()[select-1]

    if cache.is_local_channel_cache_available():
        channel_cache_unit = cache.local_channel_cache.list()[0]
        print_info(
            f"Channel: \"{channel_cache_unit['title']}\" found in cache.")

        inp = take_input('Is the channel correct?(n if no): ').lower()
        if inp in ('n', 'no'):
            # If the channel in local cache is not correct then delete it from local cache and rerun the `get_channel_id()` function
            cache.delete_local_channel_cache()
            print()
            return get_channel_id(cache, yt_resource)

    elif cache.is_shared_channel_cache_available():
        cache.shared_channel_cache.print()
        user_input = non_empty_input(
            'Select or enter channel name to search on youtube: ')

        # If the user_input is digit and in correct range then select from the given channels else search it on YouTube
        if is_int(user_input):
            # Select the channel if user_input is in correct range
            if 1 <= int(user_input) <= len(cache.shared_channel_cache):
                channel_cache_unit = cache.shared_channel_cache.list()[
                    int(user_input) - 1]

            # Display warning if user_input is out of range and rerun the function
            else:
                print_warning(
                    f'\nWrong input: Input range is [1,{len(cache.shared_channel_cache)}]')
                return get_channel_id(cache, yt_resource)
        else:
            channel_cache_unit = get_channel_from_youtube(
                user_input, yt_resource)

    else:
        user_input = non_empty_input('Search channel on youtube: ')
        channel_cache_unit = get_channel_from_youtube(
            user_input, yt_resource)

    cache.update_channel_cache(channel_cache_unit)
    cache.save_shared_channel_cache()
    return channel_cache_unit['id']


def rename(renamer: Renamer, cache: CacheManager):
    """
    Rename the files using `renamer`.

    Parameters
    ----------
    renamer: Renamer
        Contains all the information and logic required to rename files.
    cache: CacheManager
        CacheManager object to manage cache
    """

    renamer.generate_rename_dict()

    if renamer.is_rename_dict_formed():
        cache.save()
        renamer.dry_run()
        print_info('\nFiles may have been skipped if not present in YouTube playlist or already have serial number prefixed to it or the owner may have renamed the video on YouTube.\n')

        confirm = non_empty_input(
            "This can't be undone. Are you sure to rename?(y if yes): ").lower()
        print()
        if confirm in ('y', 'yes'):
            renamer.start_batch_rename()
        else:
            print("Nothing renamed")
    else:
        print_info(
            '\nFiles already have serial number or match not found in local files and youtube videos.')
        print_info(
            'Check if you have selected correct playlist.')


def main():
    yt_resource = build('youtube', 'v3', developerKey=config['api_key'])
    print_info(f"\nIf you want to ignore some files, add name of those files in {config['exceptions_file']}"
               f" in the same folder in which files to be renamed are present.")
    files = os.listdir()
    exceptions = get_exceptions()

    cache = CacheManager(config)
    if cache.is_local_playlist_cache_available():
        playlist_id = get_playlist_id_from_local_cache(cache)
        if playlist_id is None:
            # playlist_id is None if the user discarded the availaible id
            playlist_id = get_playlist_id_from_youtube(cache, yt_resource)
    else:
        playlist_id = get_playlist_id_from_youtube(cache, yt_resource)

    playlist = YTPlaylist(yt_resource, playlist_id)
    playlist.fetch_videos()

    print_heading('<< VIDEOS IN THE PLAYLIST >>')
    playlist.videos().print()
    remote_serial_dict = playlist.get_videos_serial()

    print()

    renamer = Renamer(
        remote_serial_dict, files, exceptions, character_after_serial=config['character_after_serial'], padded_zero=config['padded_zero'])
    rename(renamer, cache)


@click.command()
@click.option('-c', '--character', help='Character after serial.')
@click.option('-z', '--zero', is_flag=True, help='Add zero before serial numbers to make them all of equal length.')
@click.option('-x', '--nozero', is_flag=True, help="Don't add zero before serial numbers.")
def cli(character, zero, nozero):
    if character is not None:
        config['character_after_serial'] = character

    if zero and not nozero:
        config['padded_zero'] = True
    elif not zero and nozero:
        config['padded_zero'] = False
    # else default in config is used

    if config['api_key'] is None:
        config['api_key'] = non_empty_getpass('Please provide API key: ')
        print('You can set your api key to the environment variable YOUTUBE_DATA_API_KEY, then you will not be required to input API key everytime')

    try:
        main()
    except KeyboardInterrupt:
        print('\nAborted!')
    except Exception:
        print_warning(
            'Error occured.\nPlease check your internet connection or API key.')


if __name__ == '__main__':
    # Also set this in pyproject.toml
    # [tool.poetry.scripts]
    # ytsort = "ytsort.__main__:cli"
    cli()



import os

from ytsort.my_io import print_heading, print_warning


class Renamer:
    """
    Helps in renaming the files. This class contains the logic to rename the files according to the YouTube playlist.

    Attributes
    ----------
    rename_dict : dict[str, str]
        Dictionary as {<file's current name>: <file's new name>}
    serial_dict : dict[str,int]
        dictionary as {filename:serial}
    local_files : list[str]
        list containing files to be renamed
    exceptions : list[str]
        list of files not to be renamed
    __character_after_serial : str
        character to be put after serial number
    """

    def __init__(self, serial_dict: dict[str, int], local_files: list[str], exceptions: list[str], character_after_serial: str = '', padded_zero: bool = False):
        """
        Parameters
        ----------
        serial_dict: dict[str,int]
            dictionary containing key=filename and value=serial, i.e. {filename:serial}

        local_files: list[str]
            list containing files to be renamed

        exceptions: list[str]
            list of files not to be renamed

        character_after_serial: str
            character to be put after serial number

        padded_zero: bool
            If zero be prefixed or not to the serial numbers to make them all of equal lenghts
        """

        self.__character_after_serial = character_after_serial
        self.serial_dict = serial_dict
        self.local_files = local_files
        self.exceptions = exceptions
        self.rename_dict: dict[str, str] = {}
        if padded_zero:
            self._file_prefix = self.__padded_zero_file_prefix
            self._len_digits = len(str(len(serial_dict)))
            # self._len_digits=floor(log(len(serial_dict),10))+1
        else:
            self._file_prefix = self.__unpadded_zero_file_prefix

    @staticmethod
    def _remove_chars(st: str) -> str:
        """
        Remove the characters except alphanumerics.

        Parameters
        ----------
        st: str
            String in which characters are to be removed

        Returns
        -------
        : str
            New string after removing characters
        """

        st2 = ''
        for ch in st:
            if ch in 'qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM1234567890':
                st2 += ch
        return st2

    def _file_prefix(self, serial: int) -> str:
        """
        Prefix of file to be applied. It may be only serial, or serial padded with zero.

        [Here this method is a placeholder. It will be either assigned `__padded_zero_file_prefix` or `__unpadded_zero_file_prefix` in __init__()]
        """

        # overridden in __init__()
        return ''

    def __padded_zero_file_prefix(self, serial: int) -> str:
        """
        Return serial numbers with zero prefixed to them to make them all of equal lenghts
        """

        return f'{serial:0{self._len_digits}d}'

    def __unpadded_zero_file_prefix(self, serial: int) -> str:
        """
        Return the serial number after converting to str without prefixing zero
        """

        return str(serial)

    def _update_rename_dict(self, local_filename: str, index: int):
        """Update the `rename_dict` by adding new key-value"""
        index = self._file_prefix(index)

        self.rename_dict.update(
            {local_filename: f'{index}{self.__character_after_serial} {local_filename}'})

    def generate_rename_dict(self):
        """
        Generate dictionary to be used in `start_batch_rename()` function.
        """

        for local_file in self.local_files:
            if local_file not in self.exceptions and not os.path.isdir(local_file):
                least_length_diff = 999999
                temp_remote = None
                for remote_file in self.serial_dict:
                    if self._remove_chars(remote_file) in self._remove_chars(local_file):

                        length_diff = len(local_file) - len(remote_file)
                        if length_diff < least_length_diff:
                            least_length_diff = length_diff
                            temp_remote = remote_file

                if temp_remote is None:
                    print_warning('SKIPPED:', local_file[:75]+'...')
                elif not local_file.startswith(self._file_prefix(self.serial_dict[temp_remote])):
                    self._update_rename_dict(
                        local_file, self.serial_dict[temp_remote])

        # Sorting in arranging order according to the new name i.e. value of the dictionary
        self.rename_dict = dict(
            sorted(self.rename_dict.items(), key=lambda item: item[1]))

    def is_rename_dict_formed(self) -> bool:
        return False if len(self.rename_dict) == 0 else True

    def dry_run(self):
        """Simulate the renaming process by printing the old name and new name side by side."""

        print_heading('DRY RUN')
        u = '='*19
        print(u, 'OLD', u + '\t' + u, 'NEW', u)
        for old, new in self.rename_dict.items():
            print(old[:40]+'...', end='\t')
            print(new[:40]+'...')
            print()

    def start_batch_rename(self):
        """Rename the files according to the rename_dict"""

        for old, new in self.rename_dict.items():
            try:
                os.rename(old, new)
            except Exception as e:
                print(e)
            else:
                print(f"Done: {new[:75]}...")
